import os
import shutil
try:
    from commands import getoutput
except ImportError:
    # Python 3
    from subprocess import getoutput

from rdflib import (
    BNode,
    Graph,
    Literal,
    RDF,
    RDFS,
    URIRef,
    plugin,
)
from rdflib.store import Store


# Work in progress


def investigate_len_issue():
    store = plugin.get("SQLAlchemy", Store)(
        identifier=URIRef("rdflib_test"),
        configuration=Literal("sqlite:///%(here)s/development.sqlite" % {
                                                        "here": os.getcwd()}))
    g0 = Graph("Sleepycat")
    g0.open("/tmp/foo", create=True)
    print("Len g0 on opening: %s\n" % len(g0))
    g1 = Graph(store)
    print("Len g1 on opening: %s\n" % len(g1))
    statementId = BNode()
    print("Adding %s\n\t%s\n\t%s\n" % (statementId, RDF.type, RDF.Statement))
    g0.add((statementId, RDF.type, RDF.Statement))
    g1.add((statementId, RDF.type, RDF.Statement))
    print("Adding %s\n\t%s\n\t%s\n" % (statementId, RDF.subject,
          URIRef(u"http://rdflib.net/store/ConjunctiveGraph")))
    g0.add((statementId, RDF.subject,
           URIRef(u"http://rdflib.net/store/ConjunctiveGraph")))
    g1.add((statementId, RDF.subject,
           URIRef(u"http://rdflib.net/store/ConjunctiveGraph")))
    print("Adding %s\n\t%s\n\t%s\n" % (statementId, RDF.predicate, RDFS.label))
    g0.add((statementId, RDF.predicate, RDFS.label))
    g1.add((statementId, RDF.predicate, RDFS.label))
    print("Adding %s\n\t%s\n\t%s\n" % (
        statementId, RDF.object, Literal("Conjunctive Graph")))
    g0.add((statementId, RDF.object, Literal("Conjunctive Graph")))
    print("Len g0 after adding 4 triples %s\n" % len(g0))
    g1.add((statementId, RDF.object, Literal("Conjunctive Graph")))
    print("Len g1 after adding 4 triples %s\n" % len(g1))
    print(g0.serialize(format="nt") + "\n")
    for s, p, o in g0:
        print("s = %s\n\tp = %s\n\to = %s\n" % (
            repr(s), repr(p), repr(o)))
    print(g1.serialize(format="nt") + "\n")
    for s, p, o in g1:
        print("s = %s\n\tp = %s\n\to = %s\n" % (
            repr(s), repr(p), repr(o)))
    getoutput("cp development.sqlite devcopy.sqlite")
    print("Removing %s\n\t%s\n\t%s\n" % (statementId, RDF.type, RDF.Statement))
    g0.remove((statementId, RDF.type, RDF.Statement))
    print("Len g0 after removal %s\n" % len(g0))
    g1.remove((statementId, RDF.type, RDF.Statement))
    print("Len g1 after removal %s\n" % len(g1))
    print(g0.serialize(format="nt") + "\n")
    print(g1.serialize(format="nt") + "\n")
    g0.close()
    shutil.rmtree("/tmp/foo")
    g1.close()
    os.unlink("%(here)s/development.sqlite" % {"here": os.getcwd()})

if __name__ == "__main__":
    investigate_len_issue()



"""Convenience functions for working with Terms and Graphs."""
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef
from rdflib import Variable
from rdflib.graph import QuotedGraph
from rdflib.py3compat import format_doctest_out
from rdflib.term import Statement


__all__ = ["SUBJECT", "PREDICATE", "OBJECT", "CONTEXT", "TERM_COMBINATIONS",
           "REVERSE_TERM_COMBINATIONS", "TERM_INSTANTIATION_DICT",
           "GRAPH_TERM_DICT", "normalizeGraph", "term2Letter",
           "constructGraph", "triplePattern2termCombinations",
           "type2TermCombination", "statement2TermCombination",
           "escape_quotes"]

SUBJECT = 0
PREDICATE = 1
OBJECT = 2
CONTEXT = 3
TERM_COMBINATIONS = dict([(term, index) for index, term, in enumerate([
     "UUUU", "UUUB", "UUUF", "UUVU", "UUVB", "UUVF", "UUBU", "UUBB", "UUBF",
     "UULU", "UULB", "UULF", "UUFU", "UUFB", "UUFF",
     #
     "UVUU", "UVUB", "UVUF", "UVVU", "UVVB", "UVVF", "UVBU", "UVBB", "UVBF",
     "UVLU", "UVLB", "UVLF", "UVFU", "UVFB", "UVFF",
     #
     "VUUU", "VUUB", "VUUF", "VUVU", "VUVB", "VUVF", "VUBU", "VUBB", "VUBF",
     "VULU", "VULB", "VULF", "VUFU", "VUFB", "VUFF",
     #
     "VVUU", "VVUB", "VVUF", "VVVU", "VVVB", "VVVF", "VVBU", "VVBB", "VVBF",
     "VVLU", "VVLB", "VVLF", "VVFU", "VVFB", "VVFF",
     #
     "BUUU", "BUUB", "BUUF", "BUVU", "BUVB", "BUVF", "BUBU", "BUBB", "BUBF",
     "BULU", "BULB", "BULF", "BUFU", "BUFB", "BUFF",
     #
     "BVUU", "BVUB", "BVUF", "BVVU", "BVVB", "BVVF", "BVBU", "BVBB", "BVBF",
     "BVLU", "BVLB", "BVLF", "BVFU", "BVFB", "BVFF",
     #
     "FUUU", "FUUB", "FUUF", "FUVU", "FUVB", "FUVF", "FUBU", "FUBB", "FUBF",
     "FULU", "FULB", "FULF", "FUFU", "FUFB", "FUFF",
     #
     "FVUU", "FVUB", "FVUF", "FVVU", "FVVB", "FVVF", "FVBU", "FVBB", "FVBF",
     "FVLU", "FVLB", "FVLF", "FVFU", "FVFB", "FVFF",
     #
     # "sUUU", "sUUB", "sUUF", "sUVU", "sUVB", "sUVF", "sUBU", "sUBB", "sUBF",
     # "sULU", "sULB", "sULF", "sUFU", "sUFB", "sUFF",
     #
     # "sVUU", "sVUB", "sVUF", "sVVU", "sVVB", "sVVF", "sVBU", "sVBB", "sVBF",
     # "sVLU", "sVLB", "sVLF", "sVFU", "sVFB", "sVFF"
])])

REVERSE_TERM_COMBINATIONS = dict(
    [(value, key) for key, value in TERM_COMBINATIONS.items()])

TERM_INSTANTIATION_DICT = {
    "U": URIRef,
    "B": BNode,
    "V": Variable,
    "L": Literal
}

GRAPH_TERM_DICT = {
    "F": (QuotedGraph, URIRef),
    "U": (Graph, URIRef),
    "B": (Graph, BNode)
}


@format_doctest_out
def normalizeGraph(graph):
    """
    Take an instance of a ``Graph`` and return the instance's identifier and  ``type``.

    Types are ``U`` for a :class:`~rdflib.graph.Graph`, ``F`` for
    a :class:`~rdflib.graph.QuotedGraph` and ``B`` for a
    :class:`~rdflib.graph.ConjunctiveGraph`

    >>> from rdflib import plugin
    >>> from rdflib.graph import Graph, ConjunctiveGraph, QuotedGraph
    >>> from rdflib.store import Store
    >>> from rdflib import URIRef, Namespace
    >>> from rdflib_sqlalchemy.termutils import normalizeGraph
    >>> memstore = plugin.get('IOMemory', Store)()
    >>> g = Graph(memstore, URIRef("http://purl.org/net/bel-epa/gjh"))
    >>> normalizeGraph(g)
    (rdflib.term.URIRef(%(u)s'http://purl.org/net/bel-epa/gjh'), 'U')
    >>> g = ConjunctiveGraph(memstore, Namespace("http://rdflib.net/ns"))
    >>> normalizeGraph(g)  #doctest: +ELLIPSIS
    (rdflib.term.URIRef(%(u)s'http://rdflib.net/ns'), 'U')
    >>> g = QuotedGraph(memstore, Namespace("http://rdflib.net/ns"))
    >>> normalizeGraph(g)
    (rdflib.term.URIRef(%(u)s'http://rdflib.net/ns'), 'F')

    """
    if isinstance(graph, QuotedGraph):
        return graph.identifier, "F"
    else:
        return graph.identifier, term2Letter(graph.identifier)


@format_doctest_out
def term2Letter(term):
    """
    Relate a given term to one of several key types.

    * :class:`~rdflib.term.BNode`,
    * :class:`~rdflib.term.Literal`,
    * :class:`~rdflib.term.Statement` (Deprecated)
    * :class:`~rdflib.term.URIRef`,
    * :class:`~rdflib.term.Variable`
    * :class:`~rdflib.graph.Graph`
    * :class:`~rdflib.graph.QuotedGraph`

    >>> from rdflib import URIRef
    >>> from rdflib.term import BNode
    >>> # from rdflib.term import Statement
    >>> from rdflib.graph import Graph, QuotedGraph
    >>> from rdflib_sqlalchemy.termutils import term2Letter
    >>> term2Letter(URIRef('http://purl.org/net/bel-epa.com/'))
    'U'
    >>> term2Letter(BNode())
    'B'
    >>> term2Letter(Literal(%(u)s''))  # noqa
    'L'
    >>> term2Letter(Variable(%(u)s'x'))  # noqa
    'V'
    >>> term2Letter(Graph())
    'B'
    >>> term2Letter(QuotedGraph("IOMemory", None))
    'F'
    >>> term2Letter(None)
    'L'
    >>> # term2Letter(Statement((None, None, None), None)) # Deprecated

    """
    if isinstance(term, URIRef):
        return "U"
    elif isinstance(term, BNode):
        return "B"
    elif isinstance(term, Literal):
        return "L"
    elif isinstance(term, QuotedGraph):
        return "F"
    elif isinstance(term, Variable):
        return "V"
    elif isinstance(term, Statement):
        return "s"
    elif isinstance(term, Graph):
        return term2Letter(term.identifier)
    elif term is None:
        return "L"
    else:
        raise Exception(
            ("The given term (%s) is not an instance of any " +
             "of the known types (URIRef, BNode, Literal, QuotedGraph, " +
             "or Variable).  It is a %s")
            % (term, type(term)))


def constructGraph(key):
    """
    Return a tuple containing a ``Graph`` and an appropriate referent.

    Takes a key (one of 'F', 'U' or 'B')

    >>> from rdflib_sqlalchemy.termutils import constructGraph
    >>> constructGraph('F')
    (<class 'rdflib.graph.QuotedGraph'>, <class 'rdflib.term.URIRef'>)
    >>> constructGraph('U')
    (<class 'rdflib.graph.Graph'>, <class 'rdflib.term.URIRef'>)
    >>> constructGraph('B')
    (<class 'rdflib.graph.Graph'>, <class 'rdflib.term.BNode'>)

    """
    return GRAPH_TERM_DICT[key]


def triplePattern2termCombinations(triple):
    """Map a triple pattern to term combinations (non-functioning)."""
    s, p, o = triple
    combinations = []
    # combinations.update(TERM_COMBINATIONS)
    if isinstance(o, Literal):
        for key, val in TERM_COMBINATIONS.items():
            if key[OBJECT] == 'O':
                combinations.append(val)
    return combinations


def type2TermCombination(member, klass, context):
    """Map a type to a TermCombo."""
    try:
        rt = TERM_COMBINATIONS["%sU%s%s" %
                               (term2Letter(member),
                                term2Letter(klass),
                                normalizeGraph(context)[-1])]
        return rt
    except:
        raise Exception("Unable to persist" +
                        "classification triple: %s %s %s %s" %
                        (member, "rdf:type", klass, context))


def statement2TermCombination(subject, predicate, obj, context):
    """Map a statement to a Term Combo."""
    return TERM_COMBINATIONS["%s%s%s%s" %
                             (term2Letter(subject), term2Letter(predicate),
                              term2Letter(obj), normalizeGraph(context)[-1])]


def escape_quotes(qstr):
    """
    Escape backslashes.

    #FIXME:  This *may* prove to be a performance bottleneck and should
             perhaps be implemented in C (as it was in 4Suite RDF)

    Ported from Ft.Lib.DbUtil
    """
    if qstr is None:
        return ""
    tmp = qstr.replace("\\", "\\\\")
    tmp = tmp.replace("'", "\\'")
    return tmp



"""SQLAlchemy-based RDF store."""
from __future__ import with_statement

import hashlib
import logging
import re
import sys

import sqlalchemy
from rdflib import (
    BNode,
    Literal,
    RDF,
    URIRef
)
from rdflib.graph import Graph
from rdflib.graph import QuotedGraph
from rdflib.plugins.stores.regexmatching import PYTHON_REGEX
from rdflib.plugins.stores.regexmatching import REGEXTerm
from rdflib.store import Store
from rdflib.term import Node
from six import text_type
from six.moves import reduce
from six.moves.urllib.parse import unquote_plus
from sqlalchemy import Column, Table, MetaData, Index, types
from sqlalchemy.sql import select, expression

from .termutils import REVERSE_TERM_COMBINATIONS
from .termutils import TERM_INSTANTIATION_DICT
from .termutils import constructGraph
from .termutils import type2TermCombination
from .termutils import statement2TermCombination
from . import __version__


_logger = logging.getLogger(__name__)

COUNT_SELECT = 0
CONTEXT_SELECT = 1
TRIPLE_SELECT = 2
TRIPLE_SELECT_NO_ORDER = 3

ASSERTED_NON_TYPE_PARTITION = 3
ASSERTED_TYPE_PARTITION = 4
QUOTED_PARTITION = 5
ASSERTED_LITERAL_PARTITION = 6

FULL_TRIPLE_PARTITIONS = [QUOTED_PARTITION, ASSERTED_LITERAL_PARTITION]

INTERNED_PREFIX = "kb_"

MYSQL_MAX_INDEX_LENGTH = 200

Any = None

# Stolen from Will Waites' py4s


def skolemise(statement):
    """Skolemise."""
    def _sk(x):
        if isinstance(x, BNode):
            return URIRef("bnode:%s" % x)
        return x
    return tuple(map(_sk, statement))


def deskolemise(statement):
    """Deskolemise."""
    def _dst(x):
        if isinstance(x, URIRef) and x.startswith("bnode:"):
            _unused, bnid = x.split(":", 1)
            return BNode(bnid)
        return x
    return tuple(map(_dst, statement))


def regexp(expr, item):
    """User-defined REGEXP operator."""
    r = re.compile(expr)
    return r.match(item) is not None


def _parse_rfc1738_args(name):
    import cgi
    """ parse url str into options
    code orig from sqlalchemy.engine.url """
    pattern = re.compile(r"""
            (?P<name>[\w\+]+)://
            (?:
                (?P<username>[^:/]*)
                (?::(?P<password>[^/]*))?
            @)?
            (?:
                (?P<host>[^/:]*)
                (?::(?P<port>[^/]*))?
            )?
            (?:/(?P<database>.*))?
            """, re.X)

    m = pattern.match(name)
    if m is not None:
        (name, username, password, host, port, database) = m.group(
            1, 2, 3, 4, 5, 6)
        if database is not None:
            tokens = database.split(r"?", 2)
            database = tokens[0]
            query = (
                len(tokens) > 1 and dict(cgi.parse_qsl(tokens[1])) or None)
            if query is not None:
                query = dict([(k.encode("ascii"), query[k]) for k in query])
        else:
            query = None
        opts = {"username": username, "password": password, "host":
                host, "port": port, "database": database, "query": query}
        if opts["password"] is not None:
            opts["password"] = unquote_plus(opts["password"])
        return (name, opts)
    else:
        raise ValueError("Could not parse rfc1738 URL from string '%s'" % name)


def queryAnalysis(query, store, connection):
    """
    Helper function.

    For executing EXPLAIN on all dispatched SQL statements -
    for the pupose of analyzing index usage
    """
    res = connection.execute("explain " + query)
    rt = res.fetchall()[0]
    table, joinType, posKeys, _key, key_len, \
        comparedCol, rowsExamined, extra = rt
    if not _key:
        assert joinType == "ALL"
        if not hasattr(store, "queryOptMarks"):
            store.queryOptMarks = {}
        hits = store.queryOptMarks.get(("FULL SCAN", table), 0)
        store.queryOptMarks[("FULL SCAN", table)] = hits + 1

    if not hasattr(store, "queryOptMarks"):
        store.queryOptMarks = {}
    hits = store.queryOptMarks.get((_key, table), 0)
    store.queryOptMarks[(_key, table)] = hits + 1


def unionSELECT(selectComponents, distinct=False, selectType=TRIPLE_SELECT):
    """
    Helper function for building union all select statement.

    Terms: u - uri refs  v - variables  b - bnodes l - literal f - formula

    Takes a list of:
     - table name
     - table alias
     - table type (literal, type, asserted, quoted)
     - where clause string
    """
    selects = []
    for table, whereClause, tableType in selectComponents:

        if selectType == COUNT_SELECT:
            selectClause = table.count(whereClause)
        elif selectType == CONTEXT_SELECT:
            selectClause = expression.select([table.c.context], whereClause)
        elif tableType in FULL_TRIPLE_PARTITIONS:
            selectClause = table.select(whereClause)
        elif tableType == ASSERTED_TYPE_PARTITION:
            selectClause = expression.select(
                [table.c.id.label("id"),
                 table.c.member.label("subject"),
                 expression.literal(text_type(RDF.type)).label("predicate"),
                 table.c.klass.label("object"),
                 table.c.context.label("context"),
                 table.c.termComb.label("termcomb"),
                 expression.literal_column("NULL").label("objlanguage"),
                 expression.literal_column("NULL").label("objdatatype")][1 if __version__ <= "0.2" else 0:],
                whereClause)
        elif tableType == ASSERTED_NON_TYPE_PARTITION:
            selectClause = expression.select(
                [c for c in table.columns] +
                [expression.literal_column("NULL").label("objlanguage"),
                 expression.literal_column("NULL").label("objdatatype")],
                whereClause,
                from_obj=[table])

        selects.append(selectClause)

    orderStmt = []
    if selectType == TRIPLE_SELECT:
        orderStmt = [expression.literal_column("subject"),
                     expression.literal_column("predicate"),
                     expression.literal_column("object")]
    if distinct:
        return expression.union(*selects, **{"order_by": orderStmt})
    else:
        return expression.union_all(*selects, **{"order_by": orderStmt})


def extractTriple(tupleRt, store, hardCodedContext=None):
    """
    Extract a triple.

    Take a tuple which represents an entry in a result set and
    converts it to a tuple of terms using the termComb integer
    to interpret how to instantiate each term
    """
    if __version__ <= "0.2":
        try:
            subject, predicate, obj, rtContext, termComb, \
                objLanguage, objDatatype = tupleRt
            termCombString = REVERSE_TERM_COMBINATIONS[termComb]
            subjTerm, predTerm, objTerm, ctxTerm = termCombString
        except ValueError:
            subject, subjTerm, predicate, predTerm, obj, objTerm, \
                rtContext, ctxTerm, objLanguage, objDatatype = tupleRt
    else:
        try:
            id, subject, predicate, obj, rtContext, termComb, \
                objLanguage, objDatatype = tupleRt
            termCombString = REVERSE_TERM_COMBINATIONS[termComb]
            subjTerm, predTerm, objTerm, ctxTerm = termCombString
        except ValueError:
            id, subject, subjTerm, predicate, predTerm, obj, objTerm, \
                rtContext, ctxTerm, objLanguage, objDatatype = tupleRt

    context = rtContext is not None \
        and rtContext \
        or hardCodedContext.identifier
    s = createTerm(subject, subjTerm, store)
    p = createTerm(predicate, predTerm, store)
    o = createTerm(obj, objTerm, store, objLanguage, objDatatype)

    graphKlass, idKlass = constructGraph(ctxTerm)
    if __version__ <= "0.2":
        return s, p, o, (graphKlass, idKlass, context)
    else:
        return id, s, p, o, (graphKlass, idKlass, context)


def createTerm(
        termString, termType, store, objLanguage=None, objDatatype=None):
    # TODO: Stuff
    """
    Take a term value, term type, and store instance and creates a term object.

    QuotedGraphs are instantiated differently
    """
    if termType == "L":
        cache = store.literalCache.get((termString, objLanguage, objDatatype))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            # rt = Literal(termString, objLanguage, objDatatype)
            # store.literalCache[((termString, objLanguage, objDatatype))] = rt
            if objLanguage and not objDatatype:
                rt = Literal(termString, objLanguage)
                store.literalCache[((termString, objLanguage))] = rt
            elif objDatatype and not objLanguage:
                rt = Literal(termString, datatype=objDatatype)
                store.literalCache[((termString, objDatatype))] = rt
            elif not objLanguage and not objDatatype:
                rt = Literal(termString)
                store.literalCache[((termString))] = rt
            else:
                rt = Literal(termString, objDatatype)
                store.literalCache[((termString, objDatatype))] = rt
            return rt
    elif termType == "F":
        cache = store.otherCache.get((termType, termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = QuotedGraph(store, URIRef(termString))
            store.otherCache[(termType, termString)] = rt
            return rt
    elif termType == "B":
        cache = store.bnodeCache.get((termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = TERM_INSTANTIATION_DICT[termType](termString)
            store.bnodeCache[(termString)] = rt
            return rt
    elif termType == "U":
        cache = store.uriCache.get((termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = URIRef(termString)
            store.uriCache[(termString)] = rt
            return rt
    else:
        cache = store.otherCache.get((termType, termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = TERM_INSTANTIATION_DICT[termType](termString)
            store.otherCache[(termType, termString)] = rt
            return rt


class TermType(types.TypeDecorator):
    """Term typology."""

    impl = types.Text

    def process_bind_param(self, value, dialect):
        """Process bound parameters."""
        if isinstance(value, (QuotedGraph, Graph)):
            return text_type(value.identifier)
        elif isinstance(value, Node):
            return text_type(value)
        else:
            return value


class SQLGenerator(object):
    """SQL statement generator."""

    def buildTypeSQLCommand(self, member, klass, context):
        """Build an insert command for a type table."""
        # columns: member,klass,context
        rt = self.tables["type_statements"].insert()
        return rt, {
            "member": member,
            "klass": klass,
            "context": context.identifier,
            "termComb": int(type2TermCombination(member, klass, context))}

    def buildLiteralTripleSQLCommand(
            self, subject, predicate, obj, context):
        """
        Build an insert command for literal triples.

        (Statements where the object is a Literal).
        """
        triplePattern = int(
            statement2TermCombination(subject, predicate, obj, context))
        command = self.tables["literal_statements"].insert()
        values = {
            "subject": subject,
            "predicate": predicate,
            "object": obj,
            "context": context.identifier,
            "termComb": triplePattern,
            "objLanguage": isinstance(obj, Literal) and obj.language or None,
            "objDatatype": isinstance(obj, Literal) and obj.datatype or None
        }
        return command, values

    def buildTripleSQLCommand(
            self, subject, predicate, obj, context, quoted):
        """Build an insert command for regular triple table."""
        stmt_table = quoted and self.tables["quoted_statements"] \
            or self.tables["asserted_statements"]
        triplePattern = statement2TermCombination(
            subject, predicate, obj, context)
        command = stmt_table.insert()
        if quoted:
            params = {
                "subject": subject,
                "predicate": predicate,
                "object": obj,
                "context": context.identifier,
                "termComb": triplePattern,
                "objLanguage": isinstance(
                    obj, Literal) and obj.language or None,
                "objDatatype": isinstance(
                    obj, Literal) and obj.datatype or None
            }
        else:
            params = {
                "subject": subject,
                "predicate": predicate,
                "object": obj,
                "context": context.identifier,
                "termComb": triplePattern
            }
        return command, params

    def buildClause(
            self, table, subject, predicate, obj, context=None,
            typeTable=False):
        """Build WHERE clauses for the supplied terms and, context."""
        if typeTable:
            clauseList = [
                self.buildTypeMemberClause(subject, table),
                self.buildTypeClassClause(obj, table),
                self.buildContextClause(context, table)
            ]
        else:
            clauseList = [
                self.buildSubjClause(subject, table),
                self.buildPredClause(predicate, table),
                self.buildObjClause(obj, table),
                self.buildContextClause(context, table),
                self.buildLitDTypeClause(obj, table),
                self.buildLitLanguageClause(obj, table)
            ]

        clauseList = [clause for clause in clauseList if clause is not None]
        if clauseList:
            return expression.and_(*clauseList)
        else:
            return None

    def buildLitDTypeClause(self, obj, table):
        """Build Literal and datatype clause."""
        if isinstance(obj, Literal) and obj.datatype is not None:
            return table.c.objDatatype == obj.datatype
        else:
            return None

    def buildLitLanguageClause(self, obj, table):
        """Build Literal and language clause."""
        if isinstance(obj, Literal) and obj.language is not None:
            return table.c.objLanguage == obj.language
        else:
            return None

    # Where Clause  utility Functions
    # The predicate and object clause builders are modified in order
    # to optimize subjects and objects utility functions which can
    # take lists as their last argument (object, predicate -
    # respectively)
    def buildSubjClause(self, subject, table):
        """Build Subject clause."""
        if isinstance(subject, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.subject.op("REGEXP")(subject)
        elif isinstance(subject, list):
            # clauseStrings = [] --- unused
            return expression.or_(
                *[self.buildSubjClause(s, table) for s in subject if s])
        elif isinstance(subject, (QuotedGraph, Graph)):
            return table.c.subject == subject.identifier
        elif subject is not None:
            return table.c.subject == subject
        else:
            return None

    def buildPredClause(self, predicate, table):
        """
        Build Predicate clause.

        Capable of taking a list of predicates as well (in which case
        subclauses are joined with 'OR')
        """
        if isinstance(predicate, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.predicate.op("REGEXP")(predicate)
        elif isinstance(predicate, list):
            return expression.or_(
                *[self.buildPredClause(p, table) for p in predicate if p])
        elif predicate is not None:
            return table.c.predicate == predicate
        else:
            return None

    def buildObjClause(self, obj, table):
        """
        Build Object clause.

        Capable of taking a list of objects as well (in which case subclauses
        are joined with 'OR')
        """
        if isinstance(obj, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.object.op("REGEXP")(obj)
        elif isinstance(obj, list):
            return expression.or_(
                *[self.buildObjClause(o, table) for o in obj if o])
        elif isinstance(obj, (QuotedGraph, Graph)):
            return table.c.object == obj.identifier
        elif obj is not None:
            return table.c.object == obj
        else:
            return None

    def buildContextClause(self, context, table):
        """Build Context clause."""
        if isinstance(context, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.context.op("regexp")(context.identifier)
        elif context is not None and context.identifier is not None:
            return table.c.context == context.identifier
        else:
            return None

    def buildTypeMemberClause(self, subject, table):
        """Build Type Member clause."""
        if isinstance(subject, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.member.op("regexp")(subject)
        elif isinstance(subject, list):
            return expression.or_(
                *[self.buildTypeMemberClause(s, table) for s in subject if s])
        elif subject is not None:
            return table.c.member == subject
        else:
            return None

    def buildTypeClassClause(self, obj, table):
        """Build Type Class clause."""
        if isinstance(obj, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.klass.op("regexp")(obj)
        elif isinstance(obj, list):
            return expression.or_(
                *[self.buildTypeClassClause(o, table) for o in obj if o])
        elif obj is not None:
            return obj and table.c.klass == obj
        else:
            return None


class SQLAlchemy(Store, SQLGenerator):
    """
    SQL-92 formula-aware implementation of an rdflib Store.

    It stores its triples in the following partitions:

    - Asserted non rdf:type statements
    - Asserted literal statements
    - Asserted rdf:type statements (in a table which models Class membership)
        The motivation for this partition is primarily query speed and
        scalability as most graphs will always have more rdf:type statements
        than others
    - All Quoted statements

    In addition it persists namespace mappings in a separate table
    """

    context_aware = True
    formula_aware = True
    transaction_aware = True
    regex_matching = PYTHON_REGEX
    configuration = Literal("sqlite://")

    def __init__(self, identifier=None, configuration=None):
        """
        Initialisation.

        identifier: URIRef of the Store. Defaults to CWD
        configuration: string containing infomation open can use to
        connect to datastore.
        """
        self.identifier = identifier and identifier or "hardcoded"
        # Use only the first 10 bytes of the digest
        self._internedId = INTERNED_PREFIX + \
            hashlib.sha1(
                self.identifier.encode("utf8")).hexdigest()[:10]

        # This parameter controls how exlusively the literal table is searched
        # If true, the Literal partition is searched *exclusively* if the
        # object term in a triple pattern is a Literal or a REGEXTerm.  Note,
        # the latter case prevents the matching of URIRef nodes as the objects
        # of a triple in the store.
        # If the object term is a wildcard (None)
        # Then the Literal paritition is searched in addition to the others
        # If this parameter is false, the literal partition is searched
        # regardless of what the object of the triple pattern is
        self.STRONGLY_TYPED_TERMS = False

        self.cacheHits = 0
        self.cacheMisses = 0
        self.literalCache = {}
        self.uriCache = {}
        self.bnodeCache = {}
        self.otherCache = {}
        self.__node_pickler = None

        self.__create_table_definitions()

        if configuration is not None:
            self.configuration = configuration
            self.open(configuration)

    def __get_node_pickler(self):
        if getattr(self, "__node_pickler", False) \
                or self.__node_pickler is None:
            from rdflib.term import URIRef
            from rdflib.graph import GraphValue
            from rdflib.term import Variable
            from rdflib.term import Statement
            from rdflib.store import NodePickler
            self.__node_pickler = np = NodePickler()
            np.register(self, "S")
            np.register(URIRef, "U")
            np.register(BNode, "B")
            np.register(Literal, "L")
            np.register(Graph, "G")
            np.register(QuotedGraph, "Q")
            np.register(Variable, "V")
            np.register(Statement, "s")
            np.register(GraphValue, "v")
        return self.__node_pickler
    node_pickler = property(__get_node_pickler)

    def open(self, configuration, create=True):
        """
        Open the store specified by the configuration string.

        If create is True a store will be created if it does not already
        exist. If create is False and a store does not already exist
        an exception is raised. An exception is also raised if a store
        exists, but there is insufficient permissions to open the
        store.
        """
        name, opts = _parse_rfc1738_args(configuration)

        self.engine = sqlalchemy.create_engine(configuration)
        with self.engine.connect() as connection:
            assert connection is not None
            if create:
                self.metadata.create_all(self.engine)
        # self._db.create_function("regexp", 2, regexp)
        if configuration:
            from sqlalchemy.engine import reflection
            insp = reflection.Inspector.from_engine(self.engine)
            tbls = insp.get_table_names()
            for tn in [tbl % (self._internedId)
                       for tbl in table_name_prefixes]:
                if tn not in tbls:
                    sys.stderr.write("table %s Doesn't exist\n" % (tn))
                    # The database exists, but one of the partitions
                    # doesn't exist
                    return 0
            # Everything is there (the database and the partitions)
            return 1
        # The database doesn't exist - nothing is there
        return -1

    def close(self, commit_pending_transaction=False):
        """FIXME:  Add documentation."""
        try:
            self.engine.close()
        except:
            pass

    def destroy(self, configuration):
        """FIXME: Add documentation."""
        name, opts = _parse_rfc1738_args(configuration)
        if self.engine is None:
            # _logger.debug("Connecting in order to destroy.")
            self.engine = sqlalchemy.create_engine(configuration)
        #     _logger.debug("Connected")
        with self.engine.connect() as connection:
            trans = connection.begin()
            try:
                self.metadata.drop_all(self.engine)
                trans.commit()
            except Exception:
                e = sys.exc_info()[1]
                msg = e.args[0] if len(e.args) > 0 else ""
                _logger.debug("unable to drop table: %s " % (msg))
                trans.rollback()
        # Note, this only removes the associated tables for the closed
        # world universe given by the identifier
        # _logger.debug(
        #       "Destroyed Close World Universe %s" % (self.identifier))

    def __getBuildCommand(self, triple, context=None, quoted=False):

        subject, predicate, obj = triple
        buildCommandType = None
        if quoted or predicate != RDF.type:
            # Quoted statement or non rdf:type predicate
            # check if object is a literal
            if isinstance(obj, Literal):
                addCmd, params = self.buildLiteralTripleSQLCommand(
                    subject, predicate, obj, context)
                buildCommandType = "literal"
            else:
                addCmd, params = self.buildTripleSQLCommand(
                    subject, predicate, obj, context, quoted)
                buildCommandType = "other"
        elif predicate == RDF.type:
            # asserted rdf:type statement
            addCmd, params = self.buildTypeSQLCommand(subject, obj, context)
            buildCommandType = "type"
        return buildCommandType, addCmd, params

    # Triple Methods
    def add(self, triple, context=None, quoted=False):
        """Add a triple to the store of triples."""
        subject, predicate, obj = triple
        _, addCmd, params = self.__getBuildCommand(
            (subject, predicate, obj), context, quoted)
        with self.engine.connect() as connection:
            try:
                connection.execute(addCmd, params)
            except Exception:
                e = sys.exc_info()[1]
                msg = e.args[0] if len(e.args) > 0 else ""
                _logger.debug(
                    "Add failed %s with commands %s params %s" % (
                        msg, str(addCmd), repr(params)))
                raise

    def addN(self, quads):
        """Add a list of triples in quads form."""
        cmdTripleDict = {}

        for subject, predicate, obj, context in quads:
            buildCommandType, cmd, params = \
                self.__getBuildCommand(
                    (subject, predicate, obj),
                    context,
                    isinstance(context, QuotedGraph))

            cmdTriple = cmdTripleDict.setdefault(buildCommandType, {})
            cmdTriple.setdefault("cmd", cmd)
            cmdTriple.setdefault("params", []).append(params)

        with self.engine.connect() as connection:
            trans = connection.begin()
            try:
                for cmdTriple in cmdTripleDict.values():
                    connection.execute(cmdTriple["cmd"], cmdTriple["params"])
                trans.commit()
            except Exception:
                e = sys.exc_info()[1]
                msg = e.args[0] if len(e.args) > 0 else ""
                _logger.debug("AddN failed %s" % msg)
                trans.rollback()
                raise

    def remove(self, triple, context):
        """Remove a triple from the store."""
        subject, predicate, obj = triple
        if context is not None:
            if subject is None and predicate is None and object is None:
                self._remove_context(context)
                return
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]
        with self.engine.connect() as connection:
            trans = connection.begin()
            try:
                if not predicate or predicate != RDF.type:
                    # Need to remove predicates other than rdf:type

                    if not self.STRONGLY_TYPED_TERMS \
                            or isinstance(obj, Literal):
                        # remove literal triple
                        clause = self.buildClause(
                            literal_table, subject, predicate, obj, context)
                        connection.execute(literal_table.delete(clause))

                    for table in [quoted_table, asserted_table]:
                        # If asserted non rdf:type table and obj is Literal,
                        # don't do anything (already taken care of)
                        if table == asserted_table \
                                and isinstance(obj, Literal):
                            continue
                        else:
                            clause = self.buildClause(
                                table, subject, predicate, obj, context)
                            connection.execute(table.delete(clause))

                if predicate == RDF.type or not predicate:
                    # Need to check rdf:type and quoted partitions (in addition
                    # perhaps)
                    clause = self.buildClause(
                        asserted_type_table, subject,
                        RDF.type, obj, context, True)
                    connection.execute(asserted_type_table.delete(clause))

                    clause = self.buildClause(
                        quoted_table, subject, predicate, obj, context)
                    connection.execute(quoted_table.delete(clause))

                trans.commit()
            except Exception:
                e = sys.exc_info()[1]
                msg = e.args[0] if len(e.args) > 0 else ""
                _logger.debug("Removal failed %s" % msg)
                trans.rollback()

    def triples(self, triple, context=None):
        """
        A generator over all the triples matching pattern.

        Pattern can be any objects for comparing against nodes in
        the store, for example, RegExLiteral, Date? DateRange?

        quoted table:                <id>_quoted_statements
        asserted rdf:type table:     <id>_type_statements
        asserted non rdf:type table: <id>_asserted_statements

        triple columns: subject,predicate,object,context,termComb,
                        objLanguage,objDatatype
        class membership columns: member,klass,context termComb

        FIXME:  These union all selects *may* be further optimized by joins

        """
        subject, predicate, obj = triple

        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        if predicate == RDF.type:
            # select from asserted rdf:type partition and quoted table
            # (if a context is specified)
            typeTable = expression.alias(
                asserted_type_table, "typetable")
            clause = self.buildClause(
                typeTable, subject, RDF.type, obj, context, True)
            selects = [
                (typeTable,
                 clause,
                 ASSERTED_TYPE_PARTITION), ]

        elif isinstance(predicate, REGEXTerm) \
                and predicate.compiledExpr.match(RDF.type) \
                or not predicate:
            # Select from quoted partition (if context is specified),
            # Literal partition if (obj is Literal or None) and asserted
            # non rdf:type partition (if obj is URIRef or None)
            selects = []
            if not self.STRONGLY_TYPED_TERMS \
                    or isinstance(obj, Literal) \
                    or not obj \
                    or (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm)):
                literal = expression.alias(literal_table, "literal")
                clause = self.buildClause(
                    literal, subject, predicate, obj, context)
                selects.append((literal, clause, ASSERTED_LITERAL_PARTITION))

            if not isinstance(obj, Literal) \
                    and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS) \
                    or not obj:
                asserted = expression.alias(asserted_table, "asserted")
                clause = self.buildClause(
                    asserted, subject, predicate, obj, context)
                selects.append((asserted, clause, ASSERTED_NON_TYPE_PARTITION))

            typeTable = expression.alias(asserted_type_table, "typetable")
            clause = self.buildClause(
                typeTable, subject, RDF.type, obj, context, True)
            selects.append((typeTable, clause, ASSERTED_TYPE_PARTITION))

        elif predicate:
            # select from asserted non rdf:type partition (optionally),
            # quoted partition (if context is specified), and literal
            # partition (optionally)
            selects = []
            if not self.STRONGLY_TYPED_TERMS \
                    or isinstance(obj, Literal) \
                    or not obj \
                    or (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm)):
                literal = expression.alias(literal_table, "literal")
                clause = self.buildClause(
                    literal, subject, predicate, obj, context)
                selects.append((literal, clause, ASSERTED_LITERAL_PARTITION))

            if not isinstance(obj, Literal) \
                    and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS) \
                    or not obj:
                asserted = expression.alias(asserted_table, "asserted")
                clause = self.buildClause(
                    asserted, subject, predicate, obj, context)
                selects.append((asserted, clause, ASSERTED_NON_TYPE_PARTITION))

        if context is not None:
            quoted = expression.alias(quoted_table, "quoted")
            clause = self.buildClause(quoted, subject, predicate, obj, context)
            selects.append((quoted, clause, QUOTED_PARTITION))

        q = unionSELECT(selects, selectType=TRIPLE_SELECT_NO_ORDER)
        with self.engine.connect() as connection:
            # _logger.debug("Triples query : %s" % str(q))
            res = connection.execute(q)
            # TODO: False but it may have limitations on text column. Check
            # NOTE: SQLite does not support ORDER BY terms that aren't
            # integers, so the entire result set must be iterated in order
            # to be able to return a generator of contexts
            result = res.fetchall()
        tripleCoverage = {}
        for rt in result:
            if __version__ <= "0.2":
                s, p, o, (graphKlass, idKlass, graphId) = \
                    extractTriple(rt, self, context)
            else:
                id, s, p, o, (graphKlass, idKlass, graphId) = \
                    extractTriple(rt, self, context)
            contexts = tripleCoverage.get((s, p, o), [])
            contexts.append(graphKlass(self, idKlass(graphId)))
            tripleCoverage[(s, p, o)] = contexts

        for (s, p, o), contexts in tripleCoverage.items():
            yield (s, p, o), (c for c in contexts)

    def triples_choices(self, triple, context=None):
        """
        A variant of triples.

        Can take a list of terms instead of a single term in any slot.
        Stores can implement this to optimize the response time from the
        import default 'fallback' implementation, which will iterate over
        each term in the list and dispatch to triples.
        """
        subject, predicate, object_ = triple

        if isinstance(object_, list):
            assert not isinstance(
                subject, list), "object_ / subject are both lists"
            assert not isinstance(
                predicate, list), "object_ / predicate are both lists"
            if not object_:
                object_ = None
            for (s1, p1, o1), cg in self.triples(
                    (subject, predicate, object_), context):
                yield (s1, p1, o1), cg

        elif isinstance(subject, list):
            assert not isinstance(
                predicate, list), "subject / predicate are both lists"
            if not subject:
                subject = None
            for (s1, p1, o1), cg in self.triples(
                    (subject, predicate, object_), context):
                yield (s1, p1, o1), cg

        elif isinstance(predicate, list):
            assert not isinstance(
                subject, list), "predicate / subject are both lists"
            if not predicate:
                predicate = None
            for (s1, p1, o1), cg in self.triples(
                    (subject, predicate, object_), context):
                yield (s1, p1, o1), cg

    def __repr__(self):
        """Readable serialisation."""
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        selects = [
            (expression.alias(asserted_type_table, "typetable"),
                None, ASSERTED_TYPE_PARTITION),
            (expression.alias(quoted_table, "quoted"),
                None, QUOTED_PARTITION),
            (expression.alias(asserted_table, "asserted"),
                None, ASSERTED_NON_TYPE_PARTITION),
            (expression.alias(literal_table, "literal"),
                None, ASSERTED_LITERAL_PARTITION), ]
        q = unionSELECT(selects, distinct=False, selectType=COUNT_SELECT)
        if hasattr(self, "engine"):
            with self.engine.connect() as connection:
                res = connection.execute(q)
                rt = res.fetchall()
                typeLen, quotedLen, assertedLen, literalLen = [
                    rtTuple[0] for rtTuple in rt]
            try:
                return ("<Partitioned SQL N3 Store: %s " +
                        "contexts, %s classification assertions, " +
                        "%s quoted statements, %s property/value " +
                        "assertions, and %s other assertions>" % (
                            len([ctx for ctx in self.contexts()]),
                            typeLen, quotedLen, literalLen, assertedLen))
            except Exception:
                return "<Partitioned SQL N3 Store>"
        else:
            return "<Partitioned unopened SQL N3 Store>"

    def __len__(self, context=None):
        """Number of statements in the store."""
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        typetable = expression.alias(asserted_type_table, "typetable")
        quoted = expression.alias(quoted_table, "quoted")
        asserted = expression.alias(asserted_table, "asserted")
        literal = expression.alias(literal_table, "literal")

        quotedContext = self.buildContextClause(context, quoted)
        assertedContext = self.buildContextClause(context, asserted)
        typeContext = self.buildContextClause(context, typetable)
        literalContext = self.buildContextClause(context, literal)

        if context is not None:
            selects = [
                (typetable, typeContext,
                 ASSERTED_TYPE_PARTITION),
                (quoted, quotedContext,
                 QUOTED_PARTITION),
                (asserted, assertedContext,
                 ASSERTED_NON_TYPE_PARTITION),
                (literal, literalContext,
                 ASSERTED_LITERAL_PARTITION), ]
            q = unionSELECT(selects, distinct=True, selectType=COUNT_SELECT)
        else:
            selects = [
                (typetable, typeContext,
                 ASSERTED_TYPE_PARTITION),
                (asserted, assertedContext,
                 ASSERTED_NON_TYPE_PARTITION),
                (literal, literalContext,
                 ASSERTED_LITERAL_PARTITION), ]
            q = unionSELECT(selects, distinct=False, selectType=COUNT_SELECT)

        # _logger.debug("Length query : %s" % str(q))

        with self.engine.connect() as connection:
            res = connection.execute(q)
            rt = res.fetchall()
            # _logger.debug(rt)
            # _logger.debug(len(rt))
            return reduce(lambda x, y: x + y, [rtTuple[0] for rtTuple in rt])

    def contexts(self, triple=None):
        """Contexts."""
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        typetable = expression.alias(asserted_type_table, "typetable")
        quoted = expression.alias(quoted_table, "quoted")
        asserted = expression.alias(asserted_table, "asserted")
        literal = expression.alias(literal_table, "literal")

        if triple is not None:
            subject, predicate, obj = triple
            if predicate == RDF.type:
                # Select from asserted rdf:type partition and quoted table
                # (if a context is specified)
                clause = self.buildClause(
                    typetable, subject, RDF.type, obj, Any, True)
                selects = [(typetable, clause, ASSERTED_TYPE_PARTITION), ]

            elif isinstance(predicate, REGEXTerm) \
                    and predicate.compiledExpr.match(RDF.type) \
                    or not predicate:
                # Select from quoted partition (if context is specified),
                # literal partition if (obj is Literal or None) and
                # asserted non rdf:type partition (if obj is URIRef
                # or None)
                clause = self.buildClause(
                    typetable, subject, RDF.type, obj, Any, True)
                selects = [(typetable, clause, ASSERTED_TYPE_PARTITION), ]

                if (not self.STRONGLY_TYPED_TERMS or
                        isinstance(obj, Literal) or
                        not obj or
                        (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm))):
                    clause = self.buildClause(literal, subject, predicate, obj)
                    selects.append(
                        (literal, clause, ASSERTED_LITERAL_PARTITION))
                if not isinstance(obj, Literal) \
                        and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS) \
                        or not obj:
                    clause = self.buildClause(
                        asserted, subject, predicate, obj)
                    selects.append(
                        (asserted, clause, ASSERTED_NON_TYPE_PARTITION))

            elif predicate:
                # select from asserted non rdf:type partition (optionally),
                # quoted partition (if context is speciied), and literal
                # partition (optionally)
                selects = []
                if (not self.STRONGLY_TYPED_TERMS or
                        isinstance(obj, Literal) or
                        not obj
                        or (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm))):
                    clause = self.buildClause(
                        literal, subject, predicate, obj)
                    selects.append(
                        (literal, clause, ASSERTED_LITERAL_PARTITION))
                if not isinstance(obj, Literal) \
                        and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS) \
                        or not obj:
                    clause = self.buildClause(
                        asserted, subject, predicate, obj)
                    selects.append(
                        (asserted, clause, ASSERTED_NON_TYPE_PARTITION))

            clause = self.buildClause(quoted, subject, predicate, obj)
            selects.append((quoted, clause, QUOTED_PARTITION))
            q = unionSELECT(selects, distinct=True, selectType=CONTEXT_SELECT)
        else:
            selects = [
                (typetable, None, ASSERTED_TYPE_PARTITION),
                (quoted, None, QUOTED_PARTITION),
                (asserted, None, ASSERTED_NON_TYPE_PARTITION),
                (literal, None, ASSERTED_LITERAL_PARTITION), ]
            q = unionSELECT(selects, distinct=True, selectType=CONTEXT_SELECT)

        with self.engine.connect() as connection:
            res = connection.execute(q)
            rt = res.fetchall()
        for context in [rtTuple[0] for rtTuple in rt]:
            yield URIRef(context)

    def _remove_context(self, identifier):
        """Remove context."""
        assert identifier
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        with self.engine.connect() as connection:
            trans = connection.begin()
            try:
                for table in [quoted_table, asserted_table,
                              asserted_type_table, literal_table]:
                    clause = self.buildContextClause(identifier, table)
                    connection.execute(table.delete(clause))
                trans.commit()
            except Exception:
                e = sys.exc_info()[1]
                msg = e.args[0] if len(e.args) > 0 else ""
                _logger.debug("Context removal failed %s" % msg)
                trans.rollback()

    # Optional Namespace methods
    # Placeholder optimized interfaces (those needed in order to port Versa)
    def subjects(self, predicate=None, obj=None):
        """A generator of subjects with the given predicate and object."""
        raise Exception("Not implemented")

    # Capable of taking a list of predicate terms instead of a single term
    def objects(self, subject=None, predicate=None):
        """A generator of objects with the given subject and predicate."""
        raise Exception("Not implemented")

    # Optimized interfaces (others)
    def predicate_objects(self, subject=None):
        """A generator of (predicate, object) tuples for the given subject."""
        raise Exception("Not implemented")

    def subject_objects(self, predicate=None):
        """A generator of (subject, object) tuples for the given predicate."""
        raise Exception("Not implemented")

    def subject_predicates(self, object=None):
        """A generator of (subject, predicate) tuples for the given object."""
        raise Exception("Not implemented")

    def value(self, subject,
              predicate=u"http://www.w3.org/1999/02/22-rdf-syntax-ns#value",
              object=None, default=None, any=False):
        """
        Get a value.

        For a subject/predicate, predicate/object, or
        subject/object pair -- exactly one of subject, predicate,
        object must be None. Useful if one knows that there may only
        be one value.

        It is one of those situations that occur a lot, hence this
        'macro' like utility

        :param subject:
        :param predicate:
        :param object:  -- exactly one must be None
        :param default: -- value to be returned if no values found
        :param any: -- if true, return any value in the case there is more
                       than one, else raise a UniquenessError
        """
        raise Exception("Not implemented")

    # Namespace persistence interface implementation
    def bind(self, prefix, namespace):
        """Bind prefix for namespace."""
        with self.engine.connect() as connection:
            try:
                ins = self.tables["namespace_binds"].insert().values(
                    prefix=prefix, uri=namespace)
                connection.execute(ins)
            except Exception:
                e = sys.exc_info()[1]
                msg = e.args[0] if len(e.args) > 0 else ""
                _logger.debug("Namespace binding failed %s" % msg)

    def prefix(self, namespace):
        """Prefix."""
        with self.engine.connect() as connection:
            nb_table = self.tables["namespace_binds"]
            namespace = text_type(namespace)
            s = select([nb_table.c.prefix]).where(nb_table.c.uri == namespace)
            res = connection.execute(s)
            rt = [rtTuple[0] for rtTuple in res.fetchall()]
            res.close()
            return rt and rt[0] or None

    def namespace(self, prefix):
        """Namespace."""
        res = None
        prefix_val = text_type(prefix)
        try:
            with self.engine.connect() as connection:
                nb_table = self.tables["namespace_binds"]
                s = select([nb_table.c.uri]).where(nb_table.c.prefix == prefix_val)
                res = connection.execute(s)
                rt = [rtTuple[0] for rtTuple in res.fetchall()]
                res.close()
                # return rt and rt[0] or None
                from rdflib import URIRef
                return rt and URIRef(rt[0]) or None
        except:
            return None

    def namespaces(self):
        """Namespaces."""
        with self.engine.connect() as connection:
            res = connection.execute(self.tables["namespace_binds"].select())
            for prefix, uri in res.fetchall():
                yield prefix, uri

    def __create_table_definitions(self):
        self.metadata = MetaData()
        self.tables = {
            "asserted_statements":
            Table(
                "%s_asserted_statements" % self._internedId, self.metadata,
                Column("id", types.Integer, nullable=False, primary_key=True),
                Column("subject", TermType, nullable=False),
                Column("predicate", TermType, nullable=False),
                Column("object", TermType, nullable=False),
                Column("context", TermType, nullable=False),
                Column("termcomb", types.Integer,
                       nullable=False, key="termComb"),
                Index("%s_A_termComb_index" % self._internedId,
                      "termComb"),
                Index("%s_A_s_index" % self._internedId, "subject", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_A_p_index" % self._internedId, "predicate", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_A_o_index" % self._internedId, "object", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_A_c_index" % self._internedId, "context", mysql_length=MYSQL_MAX_INDEX_LENGTH)),
            "type_statements":
            Table("%s_type_statements" % self._internedId, self.metadata,
                  Column("id", types.Integer, nullable=False, primary_key=True),
                  Column("member", TermType, nullable=False),
                  Column("klass", TermType, nullable=False),
                  Column("context", TermType, nullable=False),
                  Column("termcomb", types.Integer, nullable=False,
                         key="termComb"),
                  Index("%s_T_termComb_index" % self._internedId,
                        "termComb"),
                  Index("%s_member_index" % self._internedId, "member", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                  Index("%s_klass_index" % self._internedId, "klass", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                  Index("%s_c_index" % self._internedId, "context", mysql_length=MYSQL_MAX_INDEX_LENGTH)),
            "literal_statements":
            Table(
                "%s_literal_statements" % self._internedId, self.metadata,
                Column("id", types.Integer, nullable=False, primary_key=True),
                Column("subject", TermType, nullable=False),
                Column("predicate", TermType, nullable=False),
                Column("object", TermType),
                Column("context", TermType, nullable=False),
                Column("termcomb", types.Integer, nullable=False,
                       key="termComb"),
                Column("objlanguage", types.String(255),
                       key="objLanguage"),
                Column("objdatatype", types.String(255),
                       key="objDatatype"),
                Index("%s_L_termComb_index" % self._internedId,
                      "termComb"),
                Index("%s_L_s_index" % self._internedId, "subject", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_L_p_index" % self._internedId, "predicate", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_L_c_index" % self._internedId, "context", mysql_length=MYSQL_MAX_INDEX_LENGTH)),
            "quoted_statements":
            Table(
                "%s_quoted_statements" % self._internedId, self.metadata,
                Column("id", types.Integer, nullable=False, primary_key=True),
                Column("subject", TermType, nullable=False),
                Column("predicate", TermType, nullable=False),
                Column("object", TermType),
                Column("context", TermType, nullable=False),
                Column("termcomb", types.Integer, nullable=False,
                       key="termComb"),
                Column("objlanguage", types.String(255),
                       key="objLanguage"),
                Column("objdatatype", types.String(255),
                       key="objDatatype"),
                Index("%s_Q_termComb_index" % self._internedId,
                      "termComb"),
                Index("%s_Q_s_index" % self._internedId, "subject", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_Q_p_index" % self._internedId, "predicate", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_Q_o_index" % self._internedId, "object", mysql_length=MYSQL_MAX_INDEX_LENGTH),
                Index("%s_Q_c_index" % self._internedId, "context", mysql_length=MYSQL_MAX_INDEX_LENGTH)),
            "namespace_binds":
            Table(
                "%s_namespace_binds" % self._internedId, self.metadata,
                Column("prefix", types.String(20), unique=True,
                       nullable=False, primary_key=True),
                Column("uri", types.Text),
                Index("%s_uri_index" % self._internedId, "uri", mysql_length=MYSQL_MAX_INDEX_LENGTH))
        }
        if __version__ > "0.2":
            for table in ["type_statements", "literal_statements",
                          "quoted_statements", "asserted_statements"]:
                self.tables[table].append_column(
                    Column("id", types.Integer, nullable=False, primary_key=True))

table_name_prefixes = [
    "%s_asserted_statements",
    "%s_type_statements",
    "%s_quoted_statements",
    "%s_namespace_binds",
    "%s_literal_statements"
]



# -*- coding: utf-8 -*-
import unittest

from rdflib import Graph
from rdflib import URIRef
from rdflib import Literal
from rdflib import plugin
from rdflib.parser import StringInputSource
from rdflib.py3compat import PY3
from rdflib.store import Store


class GraphTestCase(unittest.TestCase):
    storetest = True
    identifier = URIRef("rdflib_test")

    michel = URIRef(u"michel")
    tarek = URIRef(u"tarek")
    bob = URIRef(u"bob")
    likes = URIRef(u"likes")
    hates = URIRef(u"hates")
    pizza = URIRef(u"pizza")
    cheese = URIRef(u"cheese")

    def setUp(self, uri="sqlite://", storename=None):
        store = plugin.get(storename, Store)(identifier=self.identifier)
        self.graph = Graph(store, identifier=self.identifier)
        self.graph.open(uri, create=True)

    def tearDown(self, uri="sqlite://"):
        self.graph.destroy(uri)
        try:
            self.graph.close()
        except:
            pass

    def addStuff(self):
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        self.graph.add((tarek, likes, pizza))
        self.graph.add((tarek, likes, cheese))
        self.graph.add((michel, likes, pizza))
        self.graph.add((michel, likes, cheese))
        self.graph.add((bob, likes, cheese))
        self.graph.add((bob, hates, pizza))
        self.graph.add((bob, hates, michel))  # gasp!
        self.graph.commit()

    def removeStuff(self):
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        self.graph.remove((tarek, likes, pizza))
        self.graph.remove((tarek, likes, cheese))
        self.graph.remove((michel, likes, pizza))
        self.graph.remove((michel, likes, cheese))
        self.graph.remove((bob, likes, cheese))
        self.graph.remove((bob, hates, pizza))
        self.graph.remove((bob, hates, michel))  # gasp!

    def testAdd(self):
        self.addStuff()

    def testRemove(self):
        self.addStuff()
        self.removeStuff()

    def testTriples(self):
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese
        asserte = self.assertEquals
        triples = self.graph.triples
        Any = None

        self.addStuff()

        # unbound subjects
        asserte(len(list(triples((Any, likes, pizza)))), 2)
        asserte(len(list(triples((Any, hates, pizza)))), 1)
        asserte(len(list(triples((Any, likes, cheese)))), 3)
        asserte(len(list(triples((Any, hates, cheese)))), 0)

        # unbound objects
        asserte(len(list(triples((michel, likes, Any)))), 2)
        asserte(len(list(triples((tarek, likes, Any)))), 2)
        asserte(len(list(triples((bob, hates, Any)))), 2)
        asserte(len(list(triples((bob, likes, Any)))), 1)

        # unbound predicates
        asserte(len(list(triples((michel, Any, cheese)))), 1)
        asserte(len(list(triples((tarek, Any, cheese)))), 1)
        asserte(len(list(triples((bob, Any, pizza)))), 1)
        asserte(len(list(triples((bob, Any, michel)))), 1)

        # unbound subject, objects
        asserte(len(list(triples((Any, hates, Any)))), 2)
        asserte(len(list(triples((Any, likes, Any)))), 5)

        # unbound predicates, objects
        asserte(len(list(triples((michel, Any, Any)))), 2)
        asserte(len(list(triples((bob, Any, Any)))), 3)
        asserte(len(list(triples((tarek, Any, Any)))), 2)

        # unbound subjects, predicates
        asserte(len(list(triples((Any, Any, pizza)))), 3)
        asserte(len(list(triples((Any, Any, cheese)))), 3)
        asserte(len(list(triples((Any, Any, michel)))), 1)

        # all unbound
        asserte(len(list(triples((Any, Any, Any)))), 7)
        self.removeStuff()
        asserte(len(list(triples((Any, Any, Any)))), 0)

    def testConnected(self):
        graph = self.graph
        self.addStuff()
        self.assertEquals(True, graph.connected())

        jeroen = URIRef("jeroen")
        unconnected = URIRef("unconnected")

        graph.add((jeroen, self.likes, unconnected))

        self.assertEquals(False, graph.connected())

    def testSub(self):
        g1 = Graph()
        g2 = Graph()

        tarek = self.tarek
        # michel = self.michel
        bob = self.bob
        likes = self.likes
        # hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        g1.add((tarek, likes, pizza))
        g1.add((bob, likes, cheese))

        g2.add((bob, likes, cheese))

        g3 = g1 - g2

        self.assertEquals(len(g3), 1)
        self.assertEquals((tarek, likes, pizza) in g3, True)
        self.assertEquals((tarek, likes, cheese) in g3, False)

        self.assertEquals((bob, likes, cheese) in g3, False)

        g1 -= g2

        self.assertEquals(len(g1), 1)
        self.assertEquals((tarek, likes, pizza) in g1, True)
        self.assertEquals((tarek, likes, cheese) in g1, False)

        self.assertEquals((bob, likes, cheese) in g1, False)

    def testGraphAdd(self):
        g1 = Graph()
        g2 = Graph()

        tarek = self.tarek
        # michel = self.michel
        bob = self.bob
        likes = self.likes
        # hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        g1.add((tarek, likes, pizza))

        g2.add((bob, likes, cheese))

        g3 = g1 + g2

        self.assertEquals(len(g3), 2)
        self.assertEquals((tarek, likes, pizza) in g3, True)
        self.assertEquals((tarek, likes, cheese) in g3, False)

        self.assertEquals((bob, likes, cheese) in g3, True)

        g1 += g2

        self.assertEquals(len(g1), 2)
        self.assertEquals((tarek, likes, pizza) in g1, True)
        self.assertEquals((tarek, likes, cheese) in g1, False)

        self.assertEquals((bob, likes, cheese) in g1, True)

    def testGraphIntersection(self):
        g1 = Graph()
        g2 = Graph()

        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        # hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        g1.add((tarek, likes, pizza))
        g1.add((michel, likes, cheese))

        g2.add((bob, likes, cheese))
        g2.add((michel, likes, cheese))

        g3 = g1 * g2

        self.assertEquals(len(g3), 1)
        self.assertEquals((tarek, likes, pizza) in g3, False)
        self.assertEquals((tarek, likes, cheese) in g3, False)

        self.assertEquals((bob, likes, cheese) in g3, False)

        self.assertEquals((michel, likes, cheese) in g3, True)

        g1 *= g2

        self.assertEquals(len(g1), 1)

        self.assertEquals((tarek, likes, pizza) in g1, False)
        self.assertEquals((tarek, likes, cheese) in g1, False)

        self.assertEquals((bob, likes, cheese) in g1, False)

        self.assertEquals((michel, likes, cheese) in g1, True)

    def testStoreLiterals(self):
        bob = self.bob
        says = URIRef(u"says")
        hello = Literal(u"hello", lang="en")
        konichiwa = Literal(u"こんにちは", lang="ja")
        something = Literal(u"something")

        self.graph.add((bob, says, hello))
        self.graph.add((bob, says, konichiwa))
        self.graph.add((bob, says, something))
        self.graph.commit()

        objs = list(self.graph.objects(subject=bob, predicate=says))
        for o in objs:
            if o.value == u"hello":
                self.assertEquals(o.language, "en")
            elif o.value == u"こんにちは":
                self.assertEquals(o.language, "ja")
            elif o.value == u"something":
                self.assertEquals(o.language, None)
            else:
                self.fail()
        self.assertEquals(len(list(objs)), 3)

    def testStoreLiteralsXml(self):
        bob = self.bob
        says = URIRef(u"http://www.rdflib.net/terms/says")
        objects = [
            Literal(u"I'm the one", lang="en"),
            Literal(u"こんにちは", lang="ja"),
            Literal(u"les garçons à Noël reçoivent des œufs", lang="fr")]

        testdoc = (PY3 and bytes(xmltestdocXml, "UTF-8")) or xmltestdocXml

        self.graph.parse(StringInputSource(testdoc), formal="xml")

        objs = list(self.graph)
        self.assertEquals(len(objs), 3)
        for o in objs:
            self.assertEquals(o[0], bob)
            self.assertEquals(o[1], says)
            self.assertTrue(o[2] in objects)

    def testStoreLiteralsXmlQuote(self):
        bob = self.bob
        says = URIRef(u"http://www.rdflib.net/terms/says")
        imtheone = Literal(u"I'm the one", lang="en")

        testdoc = (PY3 and bytes(xmltestdocXmlQuote, "UTF-8")) or xmltestdocXmlQuote

        self.graph.parse(StringInputSource(testdoc), formal="xml")

        objs = list(self.graph)
        self.assertEquals(len(objs), 1)
        o = objs[0]
        self.assertEquals(o, (bob, says, imtheone))


xmltestdoc = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns="http://example.org/"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:about="http://example.org/a">
    <b rdf:resource="http://example.org/c"/>
  </rdf:Description>
</rdf:RDF>
"""

xmltestdocXml = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns:ns1="http://www.rdflib.net/terms/"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="en">I'm the one</ns1:says>
  </rdf:Description>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="ja">こんにちは</ns1:says>
  </rdf:Description>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="fr">les garçons à Noël reçoivent des œufs</ns1:says>
  </rdf:Description>
</rdf:RDF>
"""

xmltestdocXmlQuote = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns:ns1="http://www.rdflib.net/terms/"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="en">I'm the one</ns1:says>
  </rdf:Description>
</rdf:RDF>
"""


n3testdoc = """@prefix : <http://example.org/> .

:a :b :c .
"""

nttestdoc = "<http://example.org/a> <http://example.org/b> <http://example.org/c> .\n"

if __name__ == "__main__":
    unittest.main()



"""Store performance tests."""
import gc
import itertools
import os
import unittest
from random import random
from tempfile import mkdtemp
from time import time

from nose import SkipTest
from rdflib import Graph
from rdflib import URIRef
from six.moves.urllib.request import pathname2url


raise SkipTest("Store performance test suspended")


def random_uri():
    """Return random URI."""
    return URIRef("%s" % random())


class StoreTestCase(unittest.TestCase):
    """
    Test case for testing store performance.

    Probably should be something other than a unit test
    but for now we"ll add it as a unit test.
    """

    store = "IOMemory"
    path = None
    storetest = True
    performancetest = True

    def setUp(self):
        """Setup."""
        self.gcold = gc.isenabled()
        gc.collect()
        gc.disable()
        self.graph = Graph(store=self.store)
        if self.path is None:
            self.path = pathname2url(mkdtemp())
        self.graph.open(self.path)
        self.input = Graph()

    def tearDown(self):
        """Teardown."""
        self.graph.close()
        if self.gcold:
            gc.enable()
        # TODO: delete a_tmp_dir
        self.graph.close()
        del self.graph
        if hasattr(self, "path") and self.path is not None:
            if os.path.exists(self.path):
                if os.path.isdir(self.path):
                    for f in os.listdir(self.path):
                        os.unlink(self.path + "/" + f)
                    os.rmdir(self.path)
                elif len(self.path.split(":")) == 1:
                    os.unlink(self.path)
                else:
                    os.remove(self.path)

    def testTime(self):
        """Test timing."""
        # number = 1
        print("'%s': [" % self.store)
        for i in ["500triples", "1ktriples", "2ktriples",
                  "3ktriples", "5ktriples", "10ktriples",
                  "25ktriples"]:
            inputloc = os.getcwd() + "/test/sp2b/%s.n3" % i
            res = self._testInput(inputloc)
            print("%s," % res.strip())
        print("],")

    def _testInput(self, inputloc):
        number = 1
        store = self.graph
        self.input.parse(location=inputloc, format="n3")

        def add_from_input():
            for t in self.input:
                store.add(t)
        it = itertools.repeat(None, number)
        t0 = time()
        for _i in it:
            add_from_input()
        t1 = time()
        return "%.3g " % (t1 - t0)


class SQLAlchemyStoreTestCase(StoreTestCase):
    """SQLAlchemy Store."""

    store = "SQLAlchemy"

    def setUp(self):
        """Setup."""
        self.store = "SQLAlchemy"
        self.path = "sqlite:///%(here)s/test/tmpdb.sqlite" % {
            "here": os.getcwd()}
        StoreTestCase.setUp(self)

if __name__ == "__main__":
    unittest.main()



# import unittest, logging
# from rdflib import Literal
# from rdflib import plugin
# from rdflib import RDF
# from rdflib import RDFS
# from rdflib import URIRef
# from rdflib.store import Store
# from cStringIO import StringIO
# from rdflib.graph import Graph
# from rdflib.graph import ConjunctiveGraph
# from rdflib.graph import ReadOnlyGraphAggregate

# logging.basicConfig(level=logging.ERROR,format="%(message)s")
# _logger = logging.getLogger(__name__)
# _logger.setLevel(logging.DEBUG)

# testGraph1N3="""
# @prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
# @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
# @prefix : <http://test/> .
# :foo a rdfs:Class.
# :bar :d :c.
# :a :d :c.
# """

# testGraph2N3="""
# @prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
# @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
# @prefix : <http://test/> .
# @prefix log: <http://www.w3.org/2000/10/swap/log#>.
# :foo a rdfs:Resource.
# :bar rdfs:isDefinedBy [ a log:Formula ].
# :a :d :e.
# """

# testGraph3N3="""
# @prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
# @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
# @prefix log: <http://www.w3.org/2000/10/swap/log#>.
# @prefix : <http://test/> .
# <> a log:N3Document.
# """

# sparqlQ = \
# """
# PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
# SELECT *
# FROM NAMED <http://example.com/graph1>
# FROM NAMED <http://example.com/graph2>
# FROM NAMED <http://example.com/graph3>
# FROM <http://www.w3.org/2000/01/rdf-schema#>

# WHERE {?sub ?pred rdfs:Class }"""

# sparqlQ2 =\
# """
# PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
# SELECT ?class
# WHERE { GRAPH ?graph { ?member a ?class } }"""

# sparqlQ3 =\
# """
# PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
# PREFIX log: <http://www.w3.org/2000/10/swap/log#>
# SELECT ?n3Doc
# WHERE {?n3Doc a log:N3Document }"""

# # sqlalchemy_url = Literal('postgresql+psycopg2://gjh@localhost/test')
# sqlalchemy_url = Literal("mysql://gjh:50uthf0rk@localhost:3306/test")
# sqlalchemy_url = Literal("postgresql+psycopg2://gjh:50uthf0rk@localhost/test")

# class GraphAggregates1(unittest.TestCase):

#     def setUp(self):
#         memStore = plugin.get('SQLAlchemyBase', Store)(
#             identifier="rdflib_test", configuration=sqlalchemy_url)
#         self.graph1 = Graph(memStore)
#         self.graph2 = Graph(memStore)
#         self.graph3 = Graph(memStore)

#         for n3Str, graph in [(testGraph1N3, self.graph1),
#                              (testGraph2N3, self.graph2),
#                              (testGraph3N3, self.graph3)]:
#             graph.parse(StringIO(n3Str), format='n3')
#             _logger.debug("Graph... %s" % graph.serialize(format="nt"))

#         self.G = ReadOnlyGraphAggregate([self.graph1, self.graph2, self.graph3])

#     def testAggregateRaw(self):
#         #Test triples
#         assert len(list(self.G.triples((None, RDF.type, None)))) == 4,
#              len(list(self.G.triples((None, RDF.type, None))))
#         assert len(list(self.G.triples((URIRef("http://test/bar"), None, None)))) == 2
#         assert len(list(self.G.triples((None, URIRef("http://test/d"), None))))   == 3

#         #Test __len__
#         assert len(self.G) == 8, self.G.serialize(format="nt")

#         #assert context iteration
#         for g in self.G.contexts():
#             assert isinstance(g, Graph)

#         #Test __contains__
#         assert (URIRef("http://test/foo"), RDF.type, RDFS.Resource) in self.G

#         barPredicates = [URIRef("http://test/d"), RDFS.isDefinedBy]
#         assert len(list(self.G.triples_choices((URIRef("http://test/bar"), barPredicates, None)))) == 2

# class GraphAggregates2(unittest.TestCase):
#     def setUp(self):
#         memStore = plugin.get('SQLAlchemyBase', Store)(
#             identifier="rdflib_test", configuration=sqlalchemy_url)
#         self.graph1 = Graph(memStore, URIRef("http://example.com/graph1"))
#         self.graph2 = Graph(memStore, URIRef("http://example.com/graph2"))
#         self.graph3 = Graph(memStore, URIRef("http://example.com/graph3"))

#         for n3Str,graph in [(testGraph1N3, self.graph1),
#                             (testGraph2N3, self.graph2),
#                             (testGraph3N3, self.graph3)]:
#             graph.parse(StringIO(n3Str), format='n3')

#         self.graph4 = Graph(memStore, RDFS.uri)
#         self.graph4.parse(RDFS.uri)
#         self.G = ConjunctiveGraph(memStore)

#     def testAggregateSPARQL(self):
#         rt =  self.G.query(sparqlQ)
#         assert len(rt) > 1
#         rt = self.G.query(sparqlQ2, initBindings={u'?graph' : URIRef(u"http://example.com/graph3")})
#         try:
#             import json
#         except ImportError:
#             import simplejson as json
#         res = json.loads(rt.serialize(format='json'))
#         assert len(res['results']['bindings']) == 20, len(res['results']['bindings'])

# class GraphAggregates3(unittest.TestCase):
#     def setUp(self):
#         memStore = plugin.get('SQLAlchemyBase', Store)(
#             identifier="rdflib_test", configuration=sqlalchemy_url)
#         self.graph1 = Graph(memStore,URIRef("graph1"))
#         self.graph2 = Graph(memStore,URIRef("graph2"))
#         self.graph3 = Graph(memStore,URIRef("graph3"))

#         for n3Str, graph in [(testGraph1N3,self.graph1),
#                             (testGraph2N3,self.graph2),
#                             (testGraph3N3,self.graph3)]:
#             graph.parse(StringIO(n3Str),format='n3')
#         self.G = ConjunctiveGraph(memStore)

#     def testDefaultGraph(self):
#         #test that CG includes triples from all 3
#         assert self.G.query(sparqlQ3), "CG as default graph should *all* triples"
#         assert not self.graph2.query(sparqlQ3),
#             "Graph as default graph should *not* include triples from other graphs"

# if __name__ == '__main__':
#     unittest.main()



#!/usr/bin/env python
"""Setup."""
# -*- coding: utf-8 -*-
import sys
import re


def setup_python3():
    """Taken from "distribute" setup.py."""
    from distutils.filelist import FileList
    from distutils import dir_util, file_util, util
    from os.path import join, exists

    tmp_src = join("build", "src")
    # Not covered by "setup.py clean --all", so explicit deletion required.
    if exists(tmp_src):
        dir_util.remove_tree(tmp_src)
    # log.set_verbosity(1)
    fl = FileList()
    for line in open("MANIFEST.in"):
        if not line.strip():
            continue
        fl.process_template_line(line)
    dir_util.create_tree(tmp_src, fl.files)
    outfiles_2to3 = []
    for f in fl.files:
        outf, copied = file_util.copy_file(f, join(tmp_src, f), update=1)
        if copied and outf.endswith(".py"):
            outfiles_2to3.append(outf)

    util.run_2to3(outfiles_2to3)

    # arrange setup to use the copy
    sys.path.insert(0, tmp_src)

    return tmp_src

kwargs = dict(
    install_requires=[
        "rdflib>=4.0",
        "six>=1.10.0",
        "SQLAlchemy",
    ],
    tests_require="coveralls",
)

if sys.version_info[0] >= 3:
    from setuptools import setup
    # kwargs['use_2to3'] = True  # is done in setup_python3 above already
    kwargs['requires'] = []
    kwargs['src_root'] = setup_python3()
    assert setup
else:
    try:
        from setuptools import setup
        assert setup
        kwargs['test_suite'] = "nose.collector"

    except ImportError:
        from distutils.core import setup


def find_version(filename):
    """Find version.

    We have to do this because we can't import it in Python 3 until
    its been automatically converted in the setup process.
    """
    _version_re = re.compile(r'__version__ = "(.*)"')
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            return version_match.group(1)

__version__ = find_version('rdflib_sqlalchemy/__init__.py')

setup(
    name='rdflib-sqlalchemy-redux',
    version=__version__,
    description="rdflib extension adding SQLAlchemy as an AbstractSQLStore back-end store",
    author="Graham Higgins, Adam Ever-Hadani",
    author_email="gjhiggins@gmail.com, adamhadani@globality.com",
    url="http://github.com/globality-corp/rdflib-sqlalchemy",
    packages=["rdflib_sqlalchemy"],
    download_url="https://github.com/globality-corp/rdflib-sqlalchemy/zipball/master",
    license="BSD",
    platforms=["any"],
    long_description="""
    SQLAlchemy store formula-aware implementation.  It stores its triples in
    the following partitions:

    * Asserted non rdf:type statements
    * Asserted rdf:type statements (in a table which models Class membership).
      The motivation for this partition is primarily improved query speed and
      scalability as most graphs will always have more rdf:type statements than
      others.
    * All Quoted statements

    In addition it persists namespace mappings in a separate table
    """,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    entry_points={
        'rdf.plugins.store': [
            # 'SQLAlchemy = rdflib_sqlalchemy.SQLAlchemy:SQLAlchemy',
            # 'SQLAlchemyBase = rdflib_sqlalchemy.SQLAlchemyBase:SQLAlchemy',
        ],
    },
    **kwargs
)



import logging
import unittest

from rdflib import (
    BNode,
    ConjunctiveGraph,
    Literal,
    URIRef,
)
from rdflib import plugin
from rdflib.store import Store

from rdflib_sqlalchemy import registerplugins
from rdflib_sqlalchemy.store import (
    skolemise,
    deskolemise,
    _parse_rfc1738_args,
)


_logger = logging.getLogger(__name__)

michel = URIRef(u"michel")
tarek = URIRef(u"tarek")
bob = URIRef(u"bob")
likes = URIRef(u"likes")
hates = URIRef(u"hates")
pizza = URIRef(u"pizza")
cheese = URIRef(u"cheese")


class mock_cursor():
    def execute(x):
        raise Exception("Forced exception")


class SQLATestCase(unittest.TestCase):
    identifier = URIRef("rdflib_test")
    dburi = Literal("sqlite://")

    def setUp(self):
        self.store = plugin.get(
            "SQLAlchemy", Store)(identifier=self.identifier)
        self.graph = ConjunctiveGraph(self.store, identifier=self.identifier)
        self.graph.open(self.dburi, create=True)

    def tearDown(self):
        self.graph.destroy(self.dburi)
        try:
            self.graph.close()
        except:
            pass

    def test_registerplugins(self):
        # I doubt this is quite right for a fresh pip installation,
        # this test is mainly here to fill a coverage gap.
        registerplugins()
        self.assert_(plugin.get("SQLAlchemy", Store) is not None)
        p = plugin._plugins
        self.assert_(("SQLAlchemy", Store) in p, p)
        del p[("SQLAlchemy", Store)]
        plugin._plugins = p
        registerplugins()
        self.assert_(("SQLAlchemy", Store) in p, p)

    def test_skolemisation(self):
        testbnode = BNode()
        statemnt = (michel, likes, testbnode)
        res = skolemise(statemnt)
        self.assert_("bnode:N" in str(res[2]), res)

    def test_deskolemisation(self):
        testbnode = BNode()
        statemnt = (michel, likes, testbnode)
        res = deskolemise(statemnt)
        self.assert_(str(res[2]).startswith("N"), res)

    def test_redeskolemisation(self):
        testbnode = BNode()
        statemnt = skolemise((michel, likes, testbnode))
        res = deskolemise(statemnt)
        self.assert_(str(res[2]).startswith("N"), res)

    def test__parse_rfc1738_args(self):
        self.assertRaises(ValueError, _parse_rfc1738_args, "Not parseable")

    def test_namespaces(self):
        self.assert_(list(self.graph.namespaces()) != [])

    def test_contexts_without_triple(self):
        self.assert_(list(self.graph.contexts()) == [])

    def test_contexts_with_triple(self):
        statemnt = (michel, likes, pizza)
        self.assert_(self.graph.contexts(triple=statemnt) != [])

    def test__len(self):
        self.assert_(self.store.__len__() == 0)

    def test__remove_context(self):
        self.store._remove_context(self.identifier)


if __name__ == "__main__":
    unittest.main()
